#ifndef NBODY_DATAWRAPPER_HH
#define NBODY_DATAWRAPPER_HH

#include "nbody_datastructure.hh"
#include "nbody_exception.hh"


/**
 * Class template that provides unified access to the data
 * for both data structures DataVariant1 and DataVariant2.
 * (Should not be used by you as a student. Intended only
 *  for internal use by NBody_InitialSetup and NBody_VTKWriter)
 * @tparam DataVariant Type representing the data set which
 *                     contains the position, velocity and mass
 *                     of the bodies.
 */
template <typename DataVariant>
class NBody_DataWrapper
{
public:
  /**
   * Creates an instance of a DataWrapper.
   * @param[in,out] data Data set which contains the position,
   *                     velocity and mass of the bodies.
   */
  NBody_DataWrapper (DataVariant& data)
    : data_(data)
  {}

  /**
   * Access to position of the i-th body.
   */
  double& access_r (const unsigned int i, const unsigned int component)
  {
    return access_r(data_,i,component);
  }
  double access_r (const unsigned int i, const unsigned int component) const
  {
    return access_r(data_,i,component);
  }


  /**
   * Access to velocity of the i-th body.
   */
  double& access_v (const unsigned int i, const unsigned int component)
  {
    return access_v(data_,i,component);
  }
  double access_v (const unsigned int i, const unsigned int component) const
  {
    return access_v(data_,i,component);
  }

  /**
   * Access to mass of the i-th body.
   */
  double& access_m (const unsigned int i)
  {
    return access_m(data_,i);
  }
  double access_m (const unsigned int i) const
  {
    return access_m(data_,i);
  }
  
  /**
   * Get number of bodies.
   */
  unsigned int size () const
  {
    return size(data_);
  }

  /**
   * Resize data structure with a new number of bodies.
   * This method should only be needed by special initial
   * configurations with a fixed number of bodies (like
   * "threebody")
   */
  void resize (const unsigned int size)
  {
    resize(data_,size);
  }

private:
  double& access_r (DataVariant1& data,
                    const unsigned int i, const unsigned int component)
  {
    return data[i].r_i[component];
  }
  double access_r (const DataVariant1& data,
		   const unsigned int i, const unsigned int component) const
  {
    return data[i].r_i[component];
  }

  double& access_v (DataVariant1& data,
		    const  unsigned int i, const unsigned int component)
  {
    return data[i].v_i[component];
  }
  double access_v (const DataVariant1& data,
		   const  unsigned int i, const unsigned int component) const
  {
    return data[i].v_i[component];
  }

  double& access_m (DataVariant1& data, const unsigned int i)
  {
    return data[i].m_i;
  }
  double access_m (const DataVariant1& data, const unsigned int i) const
  {
    return data[i].m_i;
  }


  unsigned int size (const DataVariant1& data) const
  {
    return data.size();
  }

  void resize (DataVariant1& data, const unsigned int size)
  {
    data.resize(size);
  }

  double& access_r (DataVariant2& data,
                    const unsigned int i, const unsigned int component)
  {
    return data.r[i][component];
  }
  double access_r (const DataVariant2& data,
		   const unsigned int i, const unsigned int component) const
  {
    return data.r[i][component];
  }

  double& access_v (DataVariant2& data,
                    const unsigned int i, const unsigned int component)
  {
    return data.v[i][component];
  }
  double access_v (const DataVariant2& data,
		   const unsigned int i, const unsigned int component) const
  {
    return data.v[i][component];
  }

  double& access_m (DataVariant2& data, const unsigned int i)
  {
    return data.m[i];
  }
  double access_m (const DataVariant2& data, const unsigned int i) const
  {
    return data.m[i];
  }
  
  unsigned int size (const DataVariant2& data) const
  {
    unsigned int size_r = data.r.size();
    unsigned int size_v = data.v.size();
    unsigned int size_m = data.m.size();
    if (size_r != size_v || size_v != size_m)
      throw NBody_Exception("Sizes of vectors in DataVariant2 object differ!");
    return size_m;
  }

  void resize (DataVariant2& data, const unsigned int size)
  {
    data.r.resize(size);
    data.v.resize(size);
    data.m.resize(size);
  }

private:
  // store data reference
  DataVariant& data_;
};


#endif // NBODY_DATAWRAPPER_HH
