#ifndef NBODY_INITIALSETUP_HH
#define NBODY_INITIALSETUP_HH

#include <cmath>

#include <random>
#include <string>
#include <array>
#include <iostream>

#include "nbody_physics.hh"
#include "nbody_datawrapper.hh"
#include "nbody_exception.hh"


/**
 * Class template for choosing an initial setup, i.e. an initial condition.
 * @tparam DataVariant Type representing the data set which contains
 *                     the position, velocity and mass of the bodies.
 */
template <typename DataVariant>
class NBody_InitialSetup
{
public:
  /**
   * Creates an instance of an initial setup generator.
   * @param[in,out] data Data set which contains the position, velocity
   *                     and mass of the bodies.
   */
  NBody_InitialSetup (DataVariant& data)
    : wdata_(data)
  {}

  /**
   * Choose an initial setup.
   * @param[in] name Name of the initial setup.
   */
  void generate (const std::string& name)
  {
    if (name == "threebody")
      threebody();
    if (name == "collision")
      collision();
    // if (name == "plummer")
    //   plummer();
    if (name != "threebody" && name != "collision")
      throw NBody_Exception("Unknown initial setup " + name);
  }

private:
  /**
   * Generate initial condition for the three body problem.
   */
  void threebody ()
  {
    wdata_.resize(3);

    wdata_.access_m(0) = 3e12;
    wdata_.access_m(1) = 1.0;
    wdata_.access_m(2) = 1.0;

    wdata_.access_r(0,0) = 0.0; wdata_.access_r(0,1) = 0.0; wdata_.access_r(0,2) = 0.0;
    wdata_.access_r(1,0) = 1.0; wdata_.access_r(1,1) = 0.0; wdata_.access_r(1,2) = 0.0;
    wdata_.access_r(2,0) =-1.0; wdata_.access_r(2,1) = 0.0; wdata_.access_r(2,2) = 0.0;

    wdata_.access_v(0,0) = 0.0; wdata_.access_v(0,1) = 0.0; wdata_.access_v(0,2) = 0.0;
    wdata_.access_v(1,0) = 0.0; wdata_.access_v(1,1) = 0.1; wdata_.access_v(1,2) = 0.0;
    wdata_.access_v(2,0) = 0.0; wdata_.access_v(2,1) =-0.1; wdata_.access_v(2,2) = 0.0;
  }

  /**
   * Generate initial condition for colliding galaxies.
   */
  void collision (const int seed1 = 42, const int seed2= 24)
  {
    // initialize particles
    double ratio = 0.8;
    unsigned int b1 = 0;
    unsigned int b2 = ratio * wdata_.size();

    // initialize 1st black hole
    wdata_.access_r(b1,0) = wdata_.access_r(b1,1) = wdata_.access_r(b1,2) = 0.0;
    wdata_.access_v(b1,0) = wdata_.access_v(b1,1) = wdata_.access_v(b1,2) = 0.0;
    wdata_.access_m(b1) = 1e6; // 1 million Sonnenmassen

    generate_galaxy_around_point(seed1, {b1, b2}, 0.8*10);
    
    // initialize 2nd black hole
    wdata_.access_r(b2,0) = wdata_.access_r(b2,1) = 10.0; wdata_.access_r(b2,2) = 0.0;
    wdata_.access_m(b2) = 1e5; // 100000 Sonnenmassen

    orbitalVelocity(b1,b2);
    wdata_.access_v(b2,0) *= 0.9;
    wdata_.access_v(b2,1) *= 0.9;
    
    generate_galaxy_around_point(seed2, {b2, wdata_.size()}, 0.8*3);
  }

  void generate_galaxy_around_point(const int seed, const std::array<unsigned int,2> range, const double rad, const std::array<double,2> mass = {0.03,20})
  {
    // initialize random number generator
    std::default_random_engine gen(seed);
    std::uniform_real_distribution<double> x(-rad/2,rad/2), y(-rad/2,rad/2), z(-rad/2,rad/2);    
    std::uniform_real_distribution<double> m(mass[0], mass[1]);
      
    // initialize 1st galaxy
    for (unsigned int i = range[0] + 1; i < range[1]; ++i)
      {
	  
	wdata_.access_m(i) = m(gen);
	do{
	  wdata_.access_r(i,0) = wdata_.access_r(range[0],0) + x(gen);
	  wdata_.access_r(i,1) = wdata_.access_r(range[0],1) + y(gen);
	  wdata_.access_r(i,2) = 0;
	}while(dist(range[0],i)>rad/2);	  
	
	orbitalVelocity(range[0],i);
      }
    std::cout << std::flush;
  }

  double dist(const int p1, const int p2) const
  {
    const double x1 = wdata_.access_r(p1,0), y1 = wdata_.access_r(p1,1);
    const double x2 = wdata_.access_r(p2,0), y2 = wdata_.access_r(p2,1);

    const double r[2] = {x1-x2, y1-y2};

    return std::sqrt(r[0]*r[0] + r[1]*r[1]);    
  }
  
  /**
   * Helper function: Computes orbital velocity for a given body.
   */
  void orbitalVelocity (const int p1, const int p2)
  {
    const double m1 = wdata_.access_m(p1);
    const double r[2] = {wdata_.access_r(p1,0)-wdata_.access_r(p2,0),
			 wdata_.access_r(p1,1)-wdata_.access_r(p2,1)};
    
    // distance in parsec
    const double d = dist(p1, p2);

    // based on the distance from the sun, calculate the velocity needed to maintain a circular orbit
    const double abs_v = sqrt(G * m1 / d);

    // calculate a suitable vector perpendicular to r for the velocity of the tracer
    wdata_.access_v(p2,0) = ( r[1] / d) * abs_v;
    wdata_.access_v(p2,1) = (-r[0] / d) * abs_v;
    wdata_.access_v(p2,2) = 0.0;
  }

private:
  // store data set reference (wrapped for uniform data access)
  NBody_DataWrapper<DataVariant> wdata_;
};


#endif // NBODY_INITIALSETUP_HH
